/*
 * This file is part of vospace-ui
 * Copyright (C) 2021 Istituto Nazionale di Astrofisica
 * SPDX-License-Identifier: GPL-3.0-or-later
 */
package it.inaf.ia2.vospace.ui.controller;

import it.inaf.ia2.vospace.ui.client.VOSpaceClient;
import it.inaf.ia2.vospace.ui.data.Job;
import it.inaf.ia2.vospace.ui.exception.BadRequestException;
import java.util.List;
import java.util.stream.Collectors;
import net.ivoa.xml.uws.v1.ErrorType;
import net.ivoa.xml.uws.v1.ExecutionPhase;
import net.ivoa.xml.uws.v1.JobSummary;
import net.ivoa.xml.vospace.v2.Protocol;
import net.ivoa.xml.vospace.v2.Transfer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RestController;

@RestController
public class JobController extends BaseController {

    private static final Logger LOG = LoggerFactory.getLogger(JobController.class);
    
    @Value("${vospace-authority}")
    private String authority;

    @Autowired
    private VOSpaceClient client;

    @PostMapping(value = "/recall", consumes = MediaType.APPLICATION_JSON_VALUE, produces = MediaType.APPLICATION_JSON_VALUE)
    public ResponseEntity<Job> startRecallFromTapeJob(@RequestBody List<String> paths) {

        if (paths.isEmpty()) {
            throw new BadRequestException("Received empty list of nodes");
        }

        List<String> targets = paths.stream().map(p -> "vos://" + authority + p).collect(Collectors.toList());

        Transfer transfer = new Transfer();
        transfer.setDirection("pullToVoSpace");
        transfer.setTarget(targets);
        Protocol protocol = new Protocol();

        protocol.setUri("ia2:async-recall");
        transfer.getProtocols().add(protocol);

        JobSummary job = client.startTransferJob(transfer);
        
        if (job.getPhase() == ExecutionPhase.QUEUED
                || job.getPhase() == ExecutionPhase.PENDING
                || job.getPhase() == ExecutionPhase.EXECUTING) {
            return ResponseEntity.ok(new Job(job));
        }
        String errorMessage;
        if (job.getPhase() == ExecutionPhase.ERROR) {
            if (job.getErrorSummary() != null) {
                if (job.getErrorSummary().isHasDetail()) {
                    errorMessage = client.getErrorDetail(job.getJobId());
                } else {
                    errorMessage = job.getErrorSummary().getMessage();
                }
                if (job.getErrorSummary().getType() == ErrorType.TRANSIENT) {
                    if (!errorMessage.endsWith(".")) {
                        errorMessage += ".";
                    }
                    errorMessage += " Please retry later.";
                }
            } else {
                errorMessage = "Unexpected error";
            }
        } else {
            LOG.error("Error while executing job {}", job.getJobId());
            errorMessage = "Error while executing job. Job phase is "
                    + job.getPhase() + ". QUEUED, PENDING or EXECUTING expected";
        }
        throw new RuntimeException(errorMessage);
    }

    @GetMapping(value = "/jobs", produces = MediaType.APPLICATION_JSON_VALUE)
    public List<Job> getJobs() {
        return client.getJobs();
    }
}
